/**
 * 
  KeyCodeConverter1.ino Main sketch of "USB Keybaord code convert adapter".
  Copyright (c) 2019 okiraku-camera.tokyo
  
  This file is part of "USB Keybaord code convert adapter".

  "USB Keybaord code convert adapter" is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  "Hobo-nicola keyboard adapter" is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with "USB Keybaord code convert adapter".  If not, see <http://www.gnu.org/licenses/>.
  
*/

#include "HID.h"
#include <hidboot.h>
#include <usbhub.h>
#include <SPI.h>

// #define SERIAL_TEST 1

static const uint8_t _hidReportDescriptor[] PROGMEM = {
  0x05, 0x01, // USAGE_PAGE (Generic Desktop)
  0x09, 0x06, // USAGE (Keyboard)
  0xa1, 0x01, // COLLECTION (Application)
  0x85, 0x02, //   REPORT_ID (2) --- Mouse.cppがID==1。

  0x05, 0x07, //   USAGE_PAGE (usage = keyboard page)
    // モデファイヤキー(修飾キー)
  0x19, 0xe0, //     USAGE_MINIMUM (左CTRLが0xe0)
  0x29, 0xe7, //     USAGE_MAXIMUM (右GUIが0xe7)
  0x15, 0x00, //   LOGICAL_MINIMUM (0)
  0x25, 0x01, //   LOGICAL_MAXIMUM (1)
  0x95, 0x08, //   REPORT_COUNT (8) 全部で8つ(左右4つずつ)。
  0x75, 0x01, //   REPORT_SIZE (1) 各修飾キーにつき1ビット
  0x81, 0x02, //   INPUT (Data,Var,Abs) 8ビット長のInputフィールド(変数)が1つ。
    // 予約フィールド
  0x95, 0x01, //   REPORT_COUNT (1)
  0x75, 0x08, //   REPORT_SIZE (8) 1ビットが8つ。
  0x81, 0x01, //   INPUT (Cnst,Var,Abs)
    // LED状態のアウトプット
  0x95, 0x05, //   REPORT_COUNT (5) 全部で5つ。
  0x75, 0x01, //   REPORT_SIZE (1)  各LEDにつき1ビット
  0x05, 0x08, //   USAGE_PAGE (LEDs)
  0x19, 0x01, //     USAGE_MINIMUM (1) (NumLock LEDが1)
  0x29, 0x05, //     USAGE_MAXIMUM (5) (KANA LEDが5)
  0x91, 0x02, //   OUTPUT (Data,Var,Abs) // LED report
    // LEDレポートのパディング
  0x95, 0x01, //   REPORT_COUNT (1)
  0x75, 0x03, //   REPORT_SIZE (3)　　残りの3ビットを埋める。
  0x91, 0x01, //   OUTPUT (Cnst,Var,Abs) // padding 
    // 押下情報のインプット
  0x95, 0x06, //   REPORT_COUNT (6) 全部で6つ。
  0x75, 0x08, //   REPORT_SIZE (8) おのおの8ビットで表現
  0x15, 0x00, //   LOGICAL_MINIMUM (0) キーコードの範囲は、
  0x25, 0xdd, //   LOGICAL_MAXIMUM (221) 0～221(0xdd)まで

  0x05, 0x07, //   USAGE_PAGE (Keyboard)
  0x19, 0x00, //     USAGE_MINIMUM (0はキーコードではない)
  0x29, 0xdd, //     USAGE_MAXIMUM (Keypad Hexadecimalまで)
  0x81, 0x00, //   INPUT (Data,Ary,Abs)
  0xc0    // END_COLLECTION
};

typedef struct {
  uint16_t input;
  uint16_t output;
} table_entry_t;

#define REPORT_KEYS 6
typedef struct {
  uint8_t modifiers;
  uint8_t reserved;
  uint8_t keys[REPORT_KEYS];
} KeyReport;
KeyReport keyReport;

#if SERIAL_TEST
void dump_report(bool pressed) {
  char tmp[100];
  sprintf(tmp, "%s : %02x : %02x, %02x, %02x, %02x, %02x, %02x", 
      pressed ? "P " : "R ",
      keyReport.modifiers,
      keyReport.keys[0],keyReport.keys[1],keyReport.keys[2],keyReport.keys[3],keyReport.keys[4],keyReport.keys[5]);
  Serial.println(tmp);  
}
#endif

// KeyReportのmodifiersのビット構成
static const uint8_t MODIFIRE_LEFT_CONTROL = 0x01;
static const uint8_t MODIFIRE_LEFT_SHIFT = 0x02;
static const uint8_t MODIFIRE_LEFT_ALT = 0x04;
static const uint8_t MODIFIRE_LEFT_GUI= 0x08;
static const uint8_t MODIFIRE_RIGHT_CONTROL = 0x10;
static const uint8_t MODIFIRE_RIGHT_SHIFT = 0x20;
static const uint8_t MODIFIRE_RIGHT_ALT = 0x40;
static const uint8_t MODIFIRE_RIGHT_GUI= 0x80;

// 修飾キーの内部コード（HID仕様より）
static const uint8_t HID_LEFT_CONTROL = 0xe0;
static const uint8_t HID_LEFT_SHIFT = 0xe1;
static const uint8_t HID_LEFT_ALT = 0xe2;
static const uint8_t HID_LEFT_GUI = 0xe3;
static const uint8_t HID_RIGHT_CONTROL = 0xe4;
static const uint8_t HID_RIGHT_SHIFT = 0xe5;
static const uint8_t HID_RIGHT_ALT = 0xe6;
static const uint8_t HID_RIGHT_GUI = 0xe7;

// ふつうのキーのいくつか
static const uint8_t  HID_ENTER = 0x28;
static const uint8_t  HID_ESCAPE = 0x29;
static const uint8_t  HID_BACKSPACE = 0x2a;

static const uint8_t  HID_CAPSLOCK = 0x39;
static const uint8_t  HID_ZENHAN = 0x35;
static const uint8_t  HID_HIRAGANA = 0x88;
static const uint8_t  HID_HENKAN = 0x8a;
static const uint8_t  HID_MUHENKAN = 0x8b;

// 小指キーテスト用 （出力コード）
static const uint8_t  HID_HOME = 0x4a;
static const uint8_t  HID_PAGE_UP = 0x4b;
static const uint8_t  HID_END = 0x4d;
static const uint8_t  HID_PAGE_DOWN = 0x4e;
static const uint8_t  HID_RIGHT_ARROW = 0x4f;
static const uint8_t  HID_LEFT_ARROW = 0x50;
static const uint8_t  HID_DOWN_ARROW = 0x51;
static const uint8_t  HID_UP_ARROW = 0x52;

// 小指キーテスト用（入力コード）
static const uint8_t HID_KEY_D = 0x07;  // backspace
static const uint8_t HID_KEY_F = 0x09;  // PageDown
static const uint8_t HID_KEY_B = 0x05;  // PageUp
static const uint8_t HID_KEY_H = 0x0b;  // ←
static const uint8_t HID_KEY_J = 0x0d;  // ↓
static const uint8_t HID_KEY_K = 0x0e;  // ↑
static const uint8_t HID_KEY_L = 0x0f;  // →
static const uint8_t HID_KEY_M = 0x10;  // Enter

// テーブル転送モード開始用
static const uint8_t HID_PAUSE = 0x48;

// ローカル修飾キー（小指キー）
static const uint8_t LOCAL_MODIFIER = 0xf0;         // 小指キー内部コード
static const uint16_t WITH_LOCAL_MODIFIER = 0x100;  // 小指キー同時押し

static const uint16_t LOCAL_MODIFIER_OUTPUT = 0x100;  // 小指キー単独のとき出力許可
static const uint16_t LOCAL_MODIFIER_MAPPED_ONLY = 0x200;  // WITH_LOCAL_MODIFIER指定のないものは出力しない

static const int NUM_TABLE_ENTRIES = 50;
table_entry_t table[NUM_TABLE_ENTRIES];

static const table_entry_t test_table1[] PROGMEM = {
  {HID_CAPSLOCK, LOCAL_MODIFIER | LOCAL_MODIFIER_OUTPUT | LOCAL_MODIFIER_MAPPED_ONLY},
  {HID_ZENHAN, HID_ESCAPE},
  {HID_ESCAPE, HID_ZENHAN},
  {WITH_LOCAL_MODIFIER | HID_KEY_D, HID_BACKSPACE},
  {WITH_LOCAL_MODIFIER | HID_KEY_F, HID_PAGE_DOWN},
  {WITH_LOCAL_MODIFIER | HID_KEY_B, HID_PAGE_UP},
  {WITH_LOCAL_MODIFIER | HID_KEY_H, HID_LEFT_ARROW},
  {WITH_LOCAL_MODIFIER | HID_KEY_J, HID_DOWN_ARROW},
  {WITH_LOCAL_MODIFIER | HID_KEY_K, HID_UP_ARROW},
  {WITH_LOCAL_MODIFIER | HID_KEY_L, HID_RIGHT_ARROW},
  {WITH_LOCAL_MODIFIER | HID_KEY_M, HID_ENTER},
  {0, 0}
};

static void init_table(bool clear_data = false) {

  memcpy_P(table, test_table1, min(sizeof(test_table1), sizeof(table)) );    
}

uint8_t modifiers;
uint8_t old_modifiers;
uint8_t local_mod_state;  // 小指キー押下中
uint8_t local_mod_unused;  // 小指キー利用済 (単独打鍵検出用）
// 

void sendReport(bool pressed) {
#ifdef SERIAL_TEST
  dump_report(pressed);
#else
  HID().SendReport(2, &keyReport ,sizeof(KeyReport));
#endif  
}

void releaseAll() {
  modifiers = old_modifiers = 0;
  local_mod_state = 0;
  local_mod_unused = 0;
  memset(&keyReport, 0, sizeof(KeyReport));
  sendReport(false);
}

void report_press(uint8_t key, uint8_t mod) {
  if (key != 0) {
    bool already = false;
    int empty_slot = -1;
    for(uint8_t i = 0; i < REPORT_KEYS; i++) {
      if (keyReport.keys[i] == key)
        already = true;
      if (keyReport.keys[i] == 0 && empty_slot < 0)
        empty_slot = i;      
    }
    if (empty_slot < 0)  // error condition.
      return;
    if (!already)
      keyReport.keys[empty_slot] = key;
  }
  keyReport.modifiers = mod;
  sendReport(true);  
}

void report_release(uint8_t key, uint8_t mod) {
  if (key != 0) {
    for(uint8_t i = 0; i < REPORT_KEYS; i++) {
      if (keyReport.keys[i] == key) {
        keyReport.keys[i] = 0;
        break;
      }
    }
  }
  keyReport.modifiers = mod;
  sendReport(false);  
}

bool mapped_only = false; // 小指キーオンのときWITH_LOCAL_MOD指定のキー以外を出すかどうか

// コードの変換など。
void process_key(uint8_t key,  bool pressed, bool pending = false) {
  uint8_t key_out = key;

  if (local_mod_state && local_mod_unused && pressed)  // 小指キー押下中に何か押した
    local_mod_unused = 0;

  for(int i = 0; i < NUM_TABLE_ENTRIES; i++) {
    if (table[i].input == 0)
      break;
    if (lowByte(table[i].input) == key) { // キーが一致。
      if (lowByte(table[i].output) == LOCAL_MODIFIER) { // 小指キー検出
        if (pressed) {
          local_mod_state = 1;
          local_mod_unused = 1;  //単独打鍵検出用。修飾に使ったら0にする。
          mapped_only = (table[i].output & LOCAL_MODIFIER_MAPPED_ONLY) != 0;
        } else {
          local_mod_state = 0;
          if (keyReport.keys[0] != 0) { // modifiers以外全オフ
            memset(&keyReport, 0, sizeof(KeyReport));
            keyReport.modifiers = modifiers;
            sendReport(false);
          }
          // LOCAL_MODキーにされた元のキーの単独出力
          if ((table[i].output & LOCAL_MODIFIER_OUTPUT) && local_mod_unused) {
            report_press(key, modifiers); // keyには入力キーが入っている。修飾キーは関知せず。
            delay(20);
            report_release(key, modifiers);  // 単独打鍵出力
          }
        }
        return;  // LOCAL_MODIFIER自体の出力は無し
      }
      if (local_mod_state) {  // 小指キー押下中
        local_mod_unused = 0;  // 何か押してたら小指キー単独は無し。
        if (table[i].input & WITH_LOCAL_MODIFIER) {
          key_out = lowByte(table[i].output);
          break;
        } else if (mapped_only) // WITH_LOCAL_MOD以外のキーは出力しない。
          key_out = 0;
      } else if (highByte(table[i].input) == 0) { // ふつうの入れ替え
        key_out = lowByte(table[i].output);
        break;
      }
    }
  }
  if (key_out == 0) 
    return; // 変換先がnull。
  if (local_mod_state && mapped_only && (key == key_out))
    return; //  LOCAL_MODIFIER_MAPPED_ONLYのときは指定以外のキーを出力しない。

// modifiers
  if ( key_out >= 0xe0 && key_out <= 0xe7) {
    uint8_t mask = 1 << (key_out - 0xe0);        
    if (pressed)
      modifiers |= mask;
    else
      modifiers &= ~mask;
    key_out = 0;      
  }
  if (!pending)
    pressed ? report_press(key_out, modifiers) : report_release(key_out, modifiers);
}

class KeyboardEvent : public KeyboardReportParser {
  USBHID *phid;
public:
  KeyboardEvent() { phid = 0; };
  void Parse(USBHID *hid, bool is_rpt_id, uint8_t len, uint8_t *buf) {
    phid = hid;
    KeyboardReportParser::Parse(hid, is_rpt_id, len, buf);
  };
  void sync_led( uint8_t led ) {
    if (phid) phid->SetReport(0, 0, 2, 0, 1, &led);
  };

protected:
  void OnKeyPressed(uint8_t key) {};
  void OnKeyDown  (uint8_t mod, uint8_t key) { process_key(key, true);  };
  void OnKeyUp  (uint8_t mod, uint8_t key) { process_key(key, false); };
  void OnControlKeysChanged(uint8_t before, uint8_t after) { 
    uint8_t change = before ^ after;
    if (!change)
      return; // 変化なし
    old_modifiers = modifiers;
    for(uint8_t i = 0; i < 7; i++) {
      uint8_t mod_bit = 1 << i;
      if (change & mod_bit)
        process_key(HID_LEFT_CONTROL + i, after & mod_bit, true);
    }
    change = old_modifiers ^ modifiers;
    if (change != 0) {
      if (change & modifiers)
        report_press(0, modifiers);
      else
        report_release(0, modifiers);      
    }
  };
};

// USBハブを内蔵したキーボード（HHKとか）への対処。
USB Usb;
USBHub Hub1(&Usb);
USBHub Hub2(&Usb);
USBHub Hub3(&Usb);
USBHub Hub4(&Usb);

HIDBoot<USB_HID_PROTOCOL_KEYBOARD>    HidKeyboard(&Usb);
KeyboardEvent kbd;

void error_blink(int period) {
  for(;;) {
    TXLED1; delay(period); TXLED0; delay(period);
    RXLED1; delay(period); RXLED0; delay(period);
  }  
}

// ホストからの通知（LED）を受ける。
static uint8_t hid_led_state = 0;
static uint8_t last_hid_led_state = 0;
static uint8_t hid_idle_value = 0;
static void hid_notify(uint8_t request, uint8_t data) {
  switch(request) {
  case HID_SET_REPORT:
    hid_led_state = data;
    break;
  case HID_SET_IDLE:
    hid_idle_value = data;
    break;
  };
}

void setup() {
  init_table();
#if SERIAL_TEST  
  Serial.begin(115200);
#else  
  static HIDSubDescriptor node(_hidReportDescriptor, sizeof(_hidReportDescriptor));
  HID().AppendDescriptor(&node);
#endif

  HID().set_request_callback(hid_notify);
  delay(200);
  releaseAll();
  if (Usb.Init() == -1)
    error_blink(400);
  delay( 200 );
  HidKeyboard.SetReportParser(0, &kbd);
}

void loop() {
  Usb.Task();
  if (last_hid_led_state != hid_led_state) {
    last_hid_led_state = hid_led_state;
    kbd.sync_led(last_hid_led_state);
  }
}
